#include <stdio.h>
#include <stdlib.h>

#ifdef USE_NLLIBC
#include <nllibc.h>
#endif

#include "lib.h"
#include "type.h"
#include "asm_symbol.h"
#include "asm_code.h"
#include "asm_code_osecpu.h"

static int asm_code_osecpu_get_word_size(void)
{
  return 4;
}

static int asm_code_osecpu_get_pointer_size(void)
{
  return asm_code_osecpu_get_word_size();
}

static int asm_code_osecpu_funccall_args_reg_number(void)
{
  return 0;
}

static int asm_code_osecpu_funccall_args_stack_number(void)
{
  return VALUE_FUNCTION_ARGS_NUM;
}

static int asm_code_osecpu_tmp_reg_number(void)
{
  return 0; /* (See tmp_regs[]) */
}

static int asm_code_osecpu_tmp_stack_number(void)
{
  return VALUE_FUNCTION_ARGS_NUM;
}

static int asm_code_osecpu_function_register_number(void)
{
  return 0;
}

static int asm_code_osecpu_function_saveparam_number(void)
{
  return 0;
}

static int asm_code_osecpu_stack_align_size(void)
{
  return 1;
}

static int asm_code_osecpu_stack_correct_size(void)
{
  return 0;
}

static int bitsize(model_t model)
{
  if (model != NULL) {
    switch (model_get_size(model)) {
    case -1: ERREXIT(1);
    case  1: return 8;
    case  2: return 16;
    default: break;
    }
  }
  return 32;
}

static void asm_code_osecpu_get_value(FILE *out, long value)
{
  fprintf(out, "\tLIMM(32, R00, 0x%lx);\n", value);
}

static void asm_code_osecpu_get_value_r1(FILE *out, long value)
{
  fprintf(out, "\tLIMM(32, R01, 0x%lx);\n", value);
}

static void asm_code_osecpu_get_address_stack(FILE *out)
{
  fprintf(out, "\tCP(R00, P00);\n");
}

static void asm_code_osecpu_get_address(FILE *out, char *label)
{
  fprintf(out, "\tLIMM(32, R00, %s);\n", label);
}

static void asm_code_osecpu_add_address(FILE *out, int offset)
{
  fprintf(out, "\tLIMM(32, R02, 0x%x);\n", offset);
  fprintf(out, "\tADD(32, R00, R00, R02);\n");
}

static void asm_code_osecpu_get_r1(FILE *out)
{
  fprintf(out, "\tCP(R00, R01);\n");
}

static void asm_code_osecpu_set_r1(FILE *out)
{
  fprintf(out, "\tCP(R01, R00);\n");
}

static void asm_code_osecpu_memory_load(FILE *out, int offset, model_t model)
{
  fprintf(out, "\tLIMM(32, R02, 0x%x);\n", offset);
  fprintf(out, "\tADD(32, R02, R01, R02);\n");
  fprintf(out, "\tCP(P01, R02);\n");
  fprintf(out, "\tLMEM(%d, R00, 0, P01, 0);\n", bitsize(model));
}

static void asm_code_osecpu_memory_store(FILE *out, int offset, model_t model)
{
  fprintf(out, "\tLIMM(32, R02, 0x%x);\n", offset);
  fprintf(out, "\tADD(32, R02, R01, R02);\n");
  fprintf(out, "\tCP(P01, R02);\n");
  fprintf(out, "\tSMEM(%d, R00, 0, P01, 0);\n", bitsize(model));
}

static void asm_code_osecpu_stack_load(FILE *out, int offset, model_t model)
{
  fprintf(out, "\tCP(R03, P00);\n");
  fprintf(out, "\tLIMM(32, R02, 0x%x);\n", offset);
  fprintf(out, "\tADD(32, R02, R03, R02);\n");
  fprintf(out, "\tCP(P01, R02);\n");
  fprintf(out, "\tLMEM(%d, R00, 0, P01, 0);\n", bitsize(model));
}

static void asm_code_osecpu_stack_store(FILE *out, int offset, model_t model)
{
  fprintf(out, "\tCP(R03, P00);\n");
  fprintf(out, "\tLIMM(32, R02, 0x%x);\n", offset);
  fprintf(out, "\tADD(32, R02, R03, R02);\n");
  fprintf(out, "\tCP(P01, R02);\n");
  fprintf(out, "\tSMEM(%d, R00, 0, P01, 0);\n", bitsize(model));
}

static void asm_code_osecpu_stack_load_r1(FILE *out, int offset, model_t model)
{
  fprintf(out, "\tCP(R03, P00);\n");
  fprintf(out, "\tLIMM(32, R02, 0x%x);\n", offset);
  fprintf(out, "\tADD(32, R02, R03, R02);\n");
  fprintf(out, "\tCP(P01, R02);\n");
  fprintf(out, "\tLMEM(%d, R01, 0, P01, 0);\n", bitsize(model));
}

static void asm_code_osecpu_stack_store_r1(FILE *out, int offset, model_t model)
{
  fprintf(out, "\tCP(R03, P00);\n");
  fprintf(out, "\tLIMM(32, R02, 0x%x);\n", offset);
  fprintf(out, "\tADD(32, R02, R03, R02);\n");
  fprintf(out, "\tCP(P01, R02);\n");
  fprintf(out, "\tSMEM(%d, R01, 0, P01, 0);\n", bitsize(model));
}

static void asm_code_osecpu_stack_expand(FILE *out, int size)
{
  fprintf(out, "\tCP(R03, P00);\n");
  fprintf(out, "\tLIMM(32, R02, 0x%x);\n", size);
  fprintf(out, "\tSUB(32, R03, R03, R02);\n");
  fprintf(out, "\tCP(P00, R03);\n");
}

static void asm_code_osecpu_stack_reduce(FILE *out, int size)
{
  fprintf(out, "\tCP(R03, P00);\n");
  fprintf(out, "\tLIMM(32, R02, 0x%x);\n", size);
  fprintf(out, "\tADD(32, R03, R03, R02);\n");
  fprintf(out, "\tCP(P00, R03);\n");
}

#if 0
static char *funccall_regs[] = {
  "noreg"
};
#endif

static void asm_code_osecpu_funccall_reg_load(FILE *out, int number)
{
}

static void asm_code_osecpu_funccall_reg_store(FILE *out, int number)
{
}

#if 0
static char *tmp_regs[] = {
  "noreg",
};
#endif

static void asm_code_osecpu_tmp_reg_load(FILE *out, int number)
{
}

static void asm_code_osecpu_tmp_reg_save(FILE *out, int number)
{
}

static void asm_code_osecpu_tmp_reg_load_r1(FILE *out, int number)
{
}

static void asm_code_osecpu_tmp_reg_save_r1(FILE *out, int number)
{
}

static int asm_code_osecpu_sign_extension_char(FILE *out)
{
  return -1;
}

static int asm_code_osecpu_sign_extension_uchar(FILE *out)
{
  return -1;
}

static int asm_code_osecpu_sign_extension_short(FILE *out)
{
  return -1;
}

static int asm_code_osecpu_sign_extension_ushort(FILE *out)
{
  return -1;
}

static int asm_code_osecpu_sign_extension_int(FILE *out)
{
  return -1;
}

static int asm_code_osecpu_sign_extension_uint(FILE *out)
{
  return -1;
}

static int asm_code_osecpu_calc_inv(FILE *out, model_t model, model_t model_arg0)
{
  return -1;
}

static int asm_code_osecpu_calc_minus(FILE *out, model_t model, model_t model_arg0)
{
  return -1;
}

static int asm_code_osecpu_calc_op1(FILE *out, c_type_t type, model_t model, model_t model_arg0)
{
  return -1;
}

static void asm_code_osecpu_calc_add(FILE *out, model_t model, model_t model_arg0, model_t model_arg1)
{
  fprintf(out, "\tADD(32, R00, R00, R01);\n");
}

static void asm_code_osecpu_calc_sub(FILE *out, model_t model, model_t model_arg0, model_t model_arg1)
{
  fprintf(out, "\tSUB(32, R00, R00, R01);\n");
}

static void asm_code_osecpu_calc_and(FILE *out, model_t model, model_t model_arg0, model_t model_arg1)
{
  fprintf(out, "\tAND(32, R00, R00, R01);\n");
}

static void asm_code_osecpu_calc_or(FILE *out, model_t model, model_t model_arg0, model_t model_arg1)
{
  fprintf(out, "\tOR(32, R00, R00, R01);\n");
}

static void asm_code_osecpu_calc_xor(FILE *out, model_t model, model_t model_arg0, model_t model_arg1)
{
  fprintf(out, "\tXOR(32, R00, R00, R01);\n");
}

static int asm_code_osecpu_calc_mul(FILE *out, model_t model, model_t model_arg0, model_t model_arg1)
{
  fprintf(out, "\tMUL(32, R00, R00, R01);\n");
  return 0;
}

static int asm_code_osecpu_calc_div(FILE *out, model_t model, model_t model_arg0, model_t model_arg1)
{
  fprintf(out, "\tDIV(32, R00, R00, R01);\n");
  return 0;
}

static int asm_code_osecpu_calc_mod(FILE *out, model_t model, model_t model_arg0, model_t model_arg1)
{
  fprintf(out, "\tMOD(32, R00, R00, R01);\n");
  return 0;
}

static int asm_code_osecpu_calc_llshift(FILE *out, model_t model, model_t model_arg0, model_t model_arg1)
{
  fprintf(out, "\tSHL(32, R00, R00, R01);\n");
  return 0;
}

static int asm_code_osecpu_calc_rashift(FILE *out, model_t model, model_t model_arg0, model_t model_arg1)
{
  return -1;
}

static int asm_code_osecpu_calc_rlshift(FILE *out, model_t model, model_t model_arg0, model_t model_arg1)
{
  return -1;
}

static int asm_code_osecpu_calc_op2(FILE *out, c_type_t type, model_t model, model_t model_arg0, model_t model_arg1)
{
  return -1;
}

static void asm_code_osecpu_branch(FILE *out, char *label)
{
  fprintf(out, "\tJMP(%s);\n", label);
}

static int asm_code_osecpu_branch_zero(FILE *out, char *label)
{
  fprintf(out, "\tTSTZ(R02, R00);\n");
  fprintf(out, "\tCND(R02); JMP(%s);\n", label);
  return 0;
}

static int asm_code_osecpu_branch_nzero(FILE *out, char *label)
{
  return -1;
}

static int asm_code_osecpu_branch_cmp_eq(FILE *out, char *label, model_t model)
{
  fprintf(out, "\tCMPE(R02, R00, R01);\n");
  fprintf(out, "\tCND(R02); JMP(%s);\n", label);
  return 0;
}

static int asm_code_osecpu_branch_cmp_ne(FILE *out, char *label, model_t model)
{
  return -1;
}

static int asm_code_osecpu_branch_cmp_lt(FILE *out, char *label, model_t model)
{
  fprintf(out, "\tCMPL(R02, R00, R01);\n");
  fprintf(out, "\tCND(R02); JMP(%s);\n", label);
  return 0;
}

static int asm_code_osecpu_branch_cmp_gt(FILE *out, char *label, model_t model)
{
  return -1;
}

static int asm_code_osecpu_branch_cmp_le(FILE *out, char *label, model_t model)
{
  return -1;
}

static int asm_code_osecpu_branch_cmp_ge(FILE *out, char *label, model_t model)
{
  return -1;
}

static int asm_code_osecpu_branch_cmp_ult(FILE *out, char *label, model_t model)
{
  fprintf(out, "\tCMPL(R02, R00, R01);\n");
  fprintf(out, "\tCND(R02); JMP(%s);\n", label);
  return 0;
}

static int asm_code_osecpu_branch_cmp_ugt(FILE *out, char *label, model_t model)
{
  return -1;
}

static int asm_code_osecpu_branch_cmp_ule(FILE *out, char *label, model_t model)
{
  return -1;
}

static int asm_code_osecpu_branch_cmp_uge(FILE *out, char *label, model_t model)
{
  return -1;
}

static int asm_code_osecpu_branch_cmp(FILE *out, char *label, c_type_t type, model_t model)
{
  return -1;
}

static void asm_code_osecpu_function_call(FILE *out, char *label)
{
  fprintf(out, "\tCALL\t%s\n", label);
}

static void asm_code_osecpu_function_call_pointer(FILE *out)
{
  fprintf(out, "\tCP(P01, R00);\n");
  fprintf(out, "\tCALL(P01);\n");
}

static void asm_code_osecpu_function_start(FILE *out)
{
  fprintf(out, "\tENTER\n");
}

static void asm_code_osecpu_function_end(FILE *out)
{
  fprintf(out, "\tLEAVE\n");
  fprintf(out, "\tRET\n\n");
}

static void asm_code_osecpu_function_register_save(FILE *out)
{
}

static void asm_code_osecpu_function_register_load(FILE *out)
{
}

struct asm_code_callback asm_code_callback_osecpu = {
  asm_code_osecpu_get_word_size,
  asm_code_osecpu_get_pointer_size,

  asm_code_osecpu_funccall_args_reg_number,
  asm_code_osecpu_funccall_args_stack_number,
  asm_code_osecpu_tmp_reg_number,
  asm_code_osecpu_tmp_stack_number,
  asm_code_osecpu_function_register_number,
  asm_code_osecpu_function_saveparam_number,

  asm_code_osecpu_stack_align_size,
  asm_code_osecpu_stack_correct_size,

  asm_code_osecpu_get_value,
  asm_code_osecpu_get_value_r1,
  asm_code_osecpu_get_address_stack,
  asm_code_osecpu_get_address,
  asm_code_osecpu_add_address,
  asm_code_osecpu_get_r1,
  asm_code_osecpu_set_r1,

  asm_code_osecpu_memory_load,
  asm_code_osecpu_memory_store,
  asm_code_osecpu_stack_load,
  asm_code_osecpu_stack_store,
  asm_code_osecpu_stack_load_r1,
  asm_code_osecpu_stack_store_r1,
  asm_code_osecpu_stack_expand,
  asm_code_osecpu_stack_reduce,

  asm_code_osecpu_funccall_reg_load,
  asm_code_osecpu_funccall_reg_store,
  asm_code_osecpu_tmp_reg_load,
  asm_code_osecpu_tmp_reg_save,
  asm_code_osecpu_tmp_reg_load_r1,
  asm_code_osecpu_tmp_reg_save_r1,

  asm_code_osecpu_sign_extension_char,
  asm_code_osecpu_sign_extension_uchar,
  asm_code_osecpu_sign_extension_short,
  asm_code_osecpu_sign_extension_ushort,
  asm_code_osecpu_sign_extension_int,
  asm_code_osecpu_sign_extension_uint,

  asm_code_osecpu_calc_inv,
  asm_code_osecpu_calc_minus,

  asm_code_osecpu_calc_op1,

  asm_code_osecpu_calc_add,
  asm_code_osecpu_calc_sub,
  asm_code_osecpu_calc_and,
  asm_code_osecpu_calc_or,
  asm_code_osecpu_calc_xor,

  asm_code_osecpu_calc_mul,
  asm_code_osecpu_calc_div,
  asm_code_osecpu_calc_mod,
  asm_code_osecpu_calc_llshift,
  asm_code_osecpu_calc_rashift,
  asm_code_osecpu_calc_rlshift,

  asm_code_osecpu_calc_op2,

  asm_code_osecpu_branch,
  asm_code_osecpu_branch_zero,
  asm_code_osecpu_branch_nzero,

  asm_code_osecpu_branch_cmp_eq,
  asm_code_osecpu_branch_cmp_ne,

  asm_code_osecpu_branch_cmp_lt,
  asm_code_osecpu_branch_cmp_gt,
  asm_code_osecpu_branch_cmp_le,
  asm_code_osecpu_branch_cmp_ge,
  asm_code_osecpu_branch_cmp_ult,
  asm_code_osecpu_branch_cmp_ugt,
  asm_code_osecpu_branch_cmp_ule,
  asm_code_osecpu_branch_cmp_uge,

  asm_code_osecpu_branch_cmp,

  asm_code_osecpu_function_call,
  asm_code_osecpu_function_call_pointer,

  asm_code_osecpu_function_start,
  asm_code_osecpu_function_end,
  asm_code_osecpu_function_register_save,
  asm_code_osecpu_function_register_load,
};
